;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "minions" "1.2.0"
  "A minor-mode menu for the mode line."
  '((emacs  "26.1")
    (compat "30.1"))
  :url "https://github.com/tarsius/minions"
  :commit "5b73cd443c28a6e9c8e5ddd60ada38afdf40dfb9"
  :revdesc "v1.2.0-0-g5b73cd443c28"
  :keywords '("convenience")
  :authors '(("Jonas Bernoulli" . "emacs.minions@jonas.bernoulli.dev"))
  :maintainers '(("Jonas Bernoulli" . "emacs.minions@jonas.bernoulli.dev")))
